from typing import TypeVar, Any

from amino.state import EitherState, MaybeState

from lenses import UnboundLens
from amino import do, Do, __, Either, _, List, Map, Maybe, Lists, Just, Regex, L
from amino.json import dump_json
from amino.boolean import true
from amino.dat import Dat
from amino.lenses.lens import lens
from amino.regex import Match

from ribosome.trans.api import trans
from ribosome.plugin_state import PluginState
from ribosome.request.handler.handler import RequestHandler
from ribosome.request.handler.prefix import Full
from ribosome.trans.messages import ShowLogInfo, UpdateComponentState, Quit
from ribosome.components.scratch import Mapping
from ribosome.config import Config
from ribosome.dispatch.data import Dispatch, Internal, SendMessage, Trans

D = TypeVar('D')


@trans.free.result(trans.st)
@do(EitherState[PluginState[D], str])
def message_log() -> Do:
    yield EitherState.inspect_f(__.message_log.traverse(dump_json, Either))


@trans.free.result(trans.st)
@do(EitherState[PluginState[D], str])
def trans_log() -> Do:
    yield EitherState.inspect_f(lambda s: dump_json(s.trans_log))


@trans.free.unit(trans.st)
def set_log_level(level: str) -> None:
    handler = yield MaybeState.inspect_f(_.file_log_handler)
    handler.setLevel(level)


@trans.free.result(trans.st)
def state_data() -> str:
    return EitherState.inspect_f(lambda s: dump_json(s.data))


@trans.free.result()
def rpc_handlers() -> None:
    pass


class PatchQuery(Dat['PatchQuery']):

    def __init__(self, query: str, data: Map[str, Any]) -> None:
        self.query = query
        self.data = data


class UpdateQuery(Dat['UpdateQuery']):

    def __init__(self, patch: Maybe[PatchQuery]) -> None:
        self.patch = patch


find_rex = Regex(r'(?P<id>\w+)\((?P<key>\w+)=(?P<value>\w+)\)')


@do(Maybe[UnboundLens])
def find_lens(z: UnboundLens, m: Match) -> Do:
    id, key, value = yield m.all_groups('id', 'key', 'value').to_maybe
    yield Just(z.getattr_(id).each_().filter_(lambda a: Maybe.getattr(a, key).contains(value)))


# TODO `Regex.match_as(tpe)` that constructs a `Dat` from the `Match`
def lens_step(z: UnboundLens, s: str) -> Maybe[UnboundLens]:
    return find_rex.match(s) / L(find_lens)(z, _) | (lambda: Maybe.getattr(z, s))


def mk_lens(query: str) -> UnboundLens:
    return Lists.split(query, '.').fold_m(Just(lens))(lens_step)


@do(MaybeState[D, None])
def patch_update(query: PatchQuery) -> Do:
    lns = yield MaybeState.lift(mk_lens(query.query))
    yield MaybeState.modify(lns.modify(__.typed_copy(**query.data)))


@trans.free.unit(trans.st)
@do(MaybeState[D, None])
def update_state(query: UpdateQuery) -> Do:
    yield query.patch / patch_update | MaybeState.pure(None)


message_log_handler = RequestHandler.trans_function(message_log)(prefix=Full(), internal=true, sync=true)
trans_log_handler = RequestHandler.trans_function(trans_log)(prefix=Full(), internal=true, sync=true)
set_log_level_handler = RequestHandler.trans_function(set_log_level)(prefix=Full(), internal=true)
show_log_info_handler = RequestHandler.msg_cmd(ShowLogInfo)(prefix=Full(), internal=true)
update_state_handler = RequestHandler.trans_cmd(update_state)(json=true)
update_component_state_handler = RequestHandler.msg_cmd(UpdateComponentState)(json=true)
mapping_handler = RequestHandler.msg_fun(Mapping)(prefix=Full())
quit_handler = RequestHandler.msg_cmd(Quit)(prefix=Full())
state_handler = RequestHandler.trans_function(state_data)(name='state', internal=true, sync=true)
rpc_handlers_handler = RequestHandler.trans_function(rpc_handlers)(internal=true, sync=true, prefix=Full())


def internal_dispatchers(config: Config) -> List[Dispatch]:
    return List(
        Internal(message_log_handler),
        Internal(trans_log_handler),
        Internal(set_log_level_handler),
        SendMessage(show_log_info_handler),
        Trans(update_state_handler),
        SendMessage(update_component_state_handler),
        SendMessage(mapping_handler),
        SendMessage(quit_handler),
        Internal(state_handler),
        Internal(rpc_handlers_handler),
    )


__all__ = ('internal_dispatchers',)



from typing import TypeVar, Type, Callable, Any
from types import ModuleType

from neovim.msgpack_rpc import MsgpackStream, AsyncSession, Session
from neovim.api import Nvim
from neovim.msgpack_rpc.event_loop.base import BaseEventLoop
from neovim.msgpack_rpc.event_loop.uv import UvEventLoop
from neovim import attach

from amino import Either, _, L, Maybe, amino_log, Logger, __, Path, Map, List, Boolean, Nil, Just, env
from amino.either import ImportFailure
from amino.func import Val
from amino.logging import amino_root_file_logging
from amino.do import do, Do
from amino.dat import Dat
from amino.algebra import Algebra
from amino.boolean import false
from amino.util.string import decode
from amino.mod import instance_from_module, cls_from_module

from ribosome import NvimPlugin
from ribosome import options
from ribosome.config import Config
from ribosome.dispatch.data import Legacy, SendMessage, Trans, Internal, DispatchSync, DispatchAsync, Dispatch, DIO
from ribosome.dispatch.execute import execute_dispatch_job, request_result
from ribosome.dispatch.resolve import ComponentResolver
from ribosome.dispatch.run import DispatchJob
from ribosome.logging import ribo_log, nvim_logging
from ribosome.nvim import NvimFacade, NvimIO
from ribosome.plugin import plugin_class_from_config
from ribosome.plugin_state import PluginState, PluginStateHolder, DispatchConfig
from ribosome.request.handler.dispatcher import MsgDispatcher
from ribosome.request.handler.handler import RequestHandler
from ribosome.request.rpc import rpc_handler_functions, define_handlers, RpcHandlerSpec
from ribosome.trans.queue import PrioQueue
from ribosome.nvim.io import NS
from ribosome.trans.handler import TransComplete
from ribosome.trans.internal import internal_dispatchers

Loop = TypeVar('Loop', bound=BaseEventLoop)
NP = TypeVar('NP', bound=NvimPlugin)
D = TypeVar('D')
B = TypeVar('B')
C = TypeVar('C', bound=Config)
R = TypeVar('R')
DP = TypeVar('DP', bound=Dispatch)
RDP = TypeVar('RDP', bound=Algebra)


class HostConfig(Dat['HostConfig']):

    def __init__(
            self,
            sync_dispatch: Map[str, DispatchSync],
            async_dispatch: Map[str, DispatchAsync],
            config: DispatchConfig,
            plugin_class: Maybe[Type[NP]],
    ) -> None:
        self.sync_dispatch = sync_dispatch
        self.async_dispatch = async_dispatch
        self.config = config
        self.plugin_class = plugin_class

    @property
    def name(self) -> str:
        return self.config.name

    @property
    def dispatch(self) -> List[Dispatch]:
        return self.sync_dispatch.v + self.async_dispatch.v

    @property
    def specs(self) -> List[RpcHandlerSpec]:
        return self.dispatch / __.spec(self.name, self.config.config.prefix)

    @property
    def distinct_specs(self) -> List[RpcHandlerSpec]:
        return self.specs.distinct_by(lambda s: s.rpc_method('test'))


def dispatch_job(sync: bool,
                 dispatches: Map[str, RDP],
                 state: PluginStateHolder[D],
                 name: str,
                 prefix: str,
                 args: tuple) -> DispatchJob:
    decoded_args = decode(args)
    fun_args = decoded_args.head | Nil
    bang = decoded_args.lift(1).contains(1)
    return DispatchJob(dispatches, state, decode(name), fun_args, sync, prefix, bang)


def request_handler(vim: NvimFacade,
                    sync: bool,
                    dispatches: Map[str, RDP],
                    state: PluginStateHolder[D],
                    config: DispatchConfig) -> Callable[[str, tuple], Any]:
    sync_prefix = '' if sync else 'a'
    def handle(name: str, args: tuple) -> Do:
        try:
            job = dispatch_job(sync, dispatches, state, name, config.prefix, args)
            amino_log.debug(f'dispatching {sync_prefix}sync request: {job.name}({job.args})')
            result = request_result(job, execute_dispatch_job(job).result(vim))
            if sync:
                ribo_log.debug(f'request `{job.name}` completed: {result}')
            return vim.encode_vim_data(result)
        except Exception as e:
            amino_log.caught_exception(f'dispatching request: {name}({args})', e)
            ribo_log.error(f'fatal error dispatching request {name}({args})')
    return handle


@do(NvimIO[PluginState[D]])
def init_state(host_config: HostConfig) -> Do:
    dispatch_config = host_config.config
    data = dispatch_config.config.state()
    components = yield ComponentResolver(dispatch_config.config).run
    plugin = yield NvimIO.delay(host_config.plugin_class)
    log_handler = yield NvimIO.delay(nvim_logging)
    yield NvimIO.pure(PluginState.cons(dispatch_config, data, plugin, components, log_handler=Just(log_handler)))


# TODO can vim be injected into each request handling process?
# irrelevant if separate channel for NvimFacade is possible
@do(NvimIO[int])
def run_session(session: Session, host_config: HostConfig) -> Do:
    yield define_handlers(host_config.distinct_specs, host_config.name, host_config.name)
    state = yield init_state(host_config)
    holder = PluginStateHolder.concurrent(state)
    ribo_log.debug(f'running session with state {state}')
    yield NvimIO.delay(
        lambda vim:
        session.run(
            request_handler(vim, True, host_config.sync_dispatch, holder, host_config.config),
            request_handler(vim, False, host_config.async_dispatch, holder, host_config.config)
        )
    )
    yield NvimIO.pure(0)


def no_listen_address(err: Exception) -> None:
    raise Exception('could not connect to the vim server from within the host')


def run_loop(session: Session, prefix: str, host_config: HostConfig) -> int:
    # sock = env['NVIM_LISTEN_ADDRESS'].value_or(no_listen_address)
    # nvim = attach('socket', path=sock)
    # vim = NvimFacade(nvim, host_config.name)
    vim = NvimFacade(Nvim.from_session(session), host_config.name)
    return run_session(session, host_config).attempt(vim).get_or_raise()


def session(*args: str, loop: Loop=UvEventLoop, transport_type: str='stdio', **kwargs: str) -> Session:
    return Session(AsyncSession(MsgpackStream(loop(transport_type, *args, **kwargs))))


def start_host(prefix: str, host_config: HostConfig) -> int:
    return run_loop(session(), prefix, host_config)


def plugin_cls_from_module(mod: ModuleType) -> Either[str, Type[NP]]:
    return cls_from_module(mod, NvimPlugin)


def config_from_module(mod: ModuleType) -> Either[str, Type[C]]:
    return instance_from_module(mod, Config)


log_initialized = False


def nvim_log() -> Logger:
    global log_initialized
    if not log_initialized:
        NvimFacade.stdio_with_logging('ribosome_start_host')
        log_initialized = True
    return ribo_log


def plugin_class_dispatchers(cls: Type[NP]) -> List[DispatchSync]:
    return rpc_handler_functions(cls) / Legacy


def config_dispatchers(config: Config) -> List[DispatchAsync]:
    def choose(name: str, handler: RequestHandler) -> DispatchAsync:
        tpe = (
            SendMessage
            if isinstance(handler.dispatcher, MsgDispatcher) else
            Internal
            if handler.internal else
            Trans
        )
        return tpe(handler)
    return config.request_handlers.handlers.map2(choose)


def host_config_1(
        config: Config,
        cls: Type[NP],
        debug: Boolean,
        io_executor: Callable[[DIO], NS[PluginState[D], TransComplete]]=None,
) -> HostConfig:
    name = config.name
    cls_dispatchers = plugin_class_dispatchers(cls)
    cfg_dispatchers = config_dispatchers(config)
    int_dispatchers = internal_dispatchers(config)
    with_method = lambda ds: Map(ds.map(lambda d: (d.spec(config.name, config.prefix).rpc_method(name), d)))
    dispatches = int_dispatchers + cfg_dispatchers + cls_dispatchers
    sync_dispatch, async_dispatch = dispatches.filter(_.sync), dispatches.filter(_.async)
    dispatch_config = DispatchConfig.cons(config, io_executor)
    return HostConfig(with_method(sync_dispatch), with_method(async_dispatch), dispatch_config, cls)


class SyntheticPlugin:
    pass


def host_config(
        config: Config,
        plugin_mixin: Either[str, Type[NP]],
        debug: Boolean=false,
        io_executor: Callable[[DIO], NS[PluginState[D], TransComplete]]=None,
) -> HostConfig:
    sup = plugin_mixin | Val(SyntheticPlugin)
    cls = plugin_class_from_config(config, sup, debug)
    return host_config_1(config, cls, debug, io_executor)


def start_config_stage_2(sup: Either[str, Type[NP]], config: Config) -> int:
    debug = options.development.exists
    amino_log.debug(f'starting plugin from {config}, debug: {debug}')
    return start_host(config.name, host_config(config, sup, debug))


def error(msg: str) -> int:
    amino_log.error(msg)
    nvim_log().error(msg)
    return 1


def import_error(e: ImportFailure, desc: str) -> int:
    return error(e.expand.join_lines)


def exception(e: Exception, desc: str) -> int:
    f = __.caught_exception_error(f'starting host from {desc}', e)
    f(amino_log)
    f(nvim_log())
    return 1


def start_config_stage_1(mod: ModuleType) -> int:
    plugin_cls = plugin_cls_from_module(mod)
    return config_from_module(mod).cata(error, L(start_config_stage_2)(plugin_cls, _))


def setup_log() -> None:
    amino_root_file_logging()


def start_from(source: str, importer: Callable[[str], Either[ImportFailure, ModuleType]], desc: str) -> int:
    try:
        setup_log()
        amino_log.debug(f'start_{desc}: {source}')
        return importer(source).cata(L(import_error)(_, source), start_config_stage_1)
    except Exception as e:
        return exception(e, source)


def start_module(mod: str) -> int:
    return start_from(mod, Either.import_module, 'module')


def start_file(path: str) -> int:
    p = Path(path)
    file = p / '__init__.py' if p.is_dir() else p
    return start_from(str(file), Either.import_file, 'file')


__all__ = ('start_host', 'start_module', 'start_file')



from typing import Callable, TypeVar, Generic, Union, Any, Optional, Type

from amino import List, Nil, Maybe, Map, Either, Right, do, Do
from amino.dat import Dat
from amino.json.encoder import Encoder, json_object_with_type
from amino.json.data import JsonError, Json, JsonScalar
from amino.json.decoder import Decoder, decode

from ribosome.nvim import NvimIO
from ribosome.request.handler.handler import RequestHandler, RequestHandlers
from ribosome.config.settings import PluginSettings
from ribosome.dispatch.run import DispatchJob

A = TypeVar('A', contravariant=True)
B = TypeVar('B')
Settings = TypeVar('Settings', bound=PluginSettings)
D = TypeVar('D', bound='SimpleData')


class Config(Generic[Settings, D], Dat['Config[Settings, D]']):

    @staticmethod
    def from_opt(data: Map) -> 'Config[Settings, D]':
        return Config.cons(data.lift('name') | 'no name in json', data.lift('prefix') | None)

    @staticmethod
    def cons(
            name: str,
            prefix: Optional[str]=None,
            components: Map[str, Union[str, type]]=Map(),
            state_ctor: Optional[Callable[['Config'], D]]=None,
            settings: Optional[Settings]=None,
            request_handlers: List[RequestHandler]=Nil,
            core_components: List[str]=Nil,
            default_components: List[str]=Nil,
            request_dispatcher: Optional[Callable[[DispatchJob], NvimIO[Any]]]=None,
    ) -> 'Config[Settings, D]':
        return Config(
            name,
            prefix or name,
            components,
            state_ctor or SimpleData,
            settings or PluginSettings(name=name),
            RequestHandlers.cons(*request_handlers),
            core_components,
            default_components,
            Maybe.optional(request_dispatcher),
        )

    def __init__(
            self,
            name: str,
            prefix: str,
            components: Map[str, Union[str, type]],
            state_ctor: Callable[['Config'], D],
            settings: Settings,
            request_handlers: RequestHandlers,
            core_components: List[str],
            default_components: List[str],
            request_dispatcher: Maybe[Callable[[DispatchJob], NvimIO[Any]]],
    ) -> None:
        self.name = name
        self.prefix = prefix
        self.components = components
        self.state_ctor = state_ctor
        self.settings = settings
        self.request_handlers = request_handlers
        self.core_components = core_components
        self.default_components = default_components
        self.request_dispatcher = request_dispatcher

    def _arg_desc(self) -> List[str]:
        return List(str(self.components), str(self.settings), str(self.request_handlers))

    def state(self) -> D:
        return self.state_ctor(self)

    def vim_cmd_name(self, handler: RequestHandler) -> str:
        return handler.vim_cmd_name(self.name, self.prefix)


class Data:

    def __init__(self, config: Config) -> None:
        self.config = config

    @property
    def _str_extra(self) -> List[Any]:
        return List(self.config)

    @property
    def settings(self) -> Settings:
        return self.config.settings


class SimpleData(Dat['SimpleData'], Data):

    def __init__(self, config: Config) -> None:
        super().__init__(config)


class ConfigEncoder(Encoder[Config], tpe=Config):

    def encode(self, a: Config) -> Either[JsonError, Map]:
        return Right(json_object_with_type(Map(name=JsonScalar(a.name)), Config))


class ConfigDecoder(Decoder[Config], tpe=Config):

    @do(Either[JsonError, Config])
    def decode(self, tpe: Type[Config], data: Json) -> Do:
        f = yield data.field('name')
        name = yield decode(f)
        yield Right(Config.cons(name))


__all__ = ('Config', 'Data', 'SimpleData')



import abc
import inspect
import logging
from typing import TypeVar, Generic, Type, Any, Callable, Optional
from threading import Lock

import greenlet

import toolz

from amino import Map, List, Boolean, Nil, Either, _, Lists, Maybe, Nothing, Try, do, Do, IO
from amino.dat import Dat
from amino.func import flip

from ribosome.trans.message_base import Message, Sendable, Envelope
from ribosome.dispatch.component import Component
from ribosome.nvim.io import NvimIOState, NS
from ribosome.dispatch.data import DispatchResult, DIO, Dispatch
from ribosome.trans.queue import PrioQueue
from ribosome.trans.legacy import Handler
from ribosome.trans.handler import TransHandler, TransComplete
from ribosome.nvim import NvimIO
from ribosome.logging import Logging

D = TypeVar('D')
TransState = NvimIOState[D, DispatchResult]


class Handlers(Dat['Handlers']):

    def __init__(self, prio: int, handlers: Map[type, Handler]) -> None:
        self.prio = prio
        self.handlers = handlers

    def handler(self, msg):
        return self.handlers.get(type(msg))


def handlers(cls: Type['MachineBase']) -> List[Handler]:
    return Lists.wrap(inspect.getmembers(cls, Boolean.is_a(TransHandler))) / _[1]


def message_handlers(handlers: List[Handler]) -> Map[float, Handlers]:
    def create(prio, h):
        h = List.wrap(h).apzip(_.message).map2(flip)
        return prio, Handlers(prio, Map(h))
    return Map(toolz.groupby(_.prio, handlers)).map(create)


class ComponentState(Dat['ComponentState']):

    @staticmethod
    def cons(comp: Component) -> 'ComponentState':
        hs = message_handlers(handlers(comp))
        return ComponentState(comp, hs)

    def __init__(self, component: Component, handlers: Map[float, Handlers]) -> None:
        self.component = component
        self.handlers = handlers

    @property
    def name(self) -> str:
        return self.component.name


class Components(Dat['Components']):

    def __init__(self, all: List[ComponentState]) -> None:
        self.all = all

    def by_name(self, name: str) -> Either[str, ComponentState]:
        return self.all.find(_.name == name).to_either(f'no component named {name}')


class DispatchConfig(Dat['DispatchConfig']):

    @staticmethod
    def cons(
            config: 'ribosome.config.Config',
            io_executor: Callable[[DIO], NS['PluginState[D]', TransComplete]]=None,
    ) -> 'DispatchConfig':
        return DispatchConfig(config, Maybe.optional(io_executor))

    def __init__(
            self,
            config: 'ribosome.config.Config',
            io_executor: Maybe[Callable[[DIO], NS['PluginState[D]', TransComplete]]]
    ) -> None:
        self.config = config
        self.io_executor = io_executor

    @property
    def name(self) -> str:
        return self.config.name

    @property
    def prefix(self) -> str:
        return self.config.prefix


class PluginState(Generic[D], Dat['PluginState']):

    @staticmethod
    def cons(
            dispatch_config: DispatchConfig,
            data: D,
            plugin: Any,
            components: List[Component],
            messages: PrioQueue[Message]=PrioQueue.empty,
            message_log: List[Message]=Nil,
            trans_log: List[str]=Nil,
            log_handler: Maybe[logging.Handler]=Nothing,
    ) -> 'PluginState':
        component_state = Components(components / ComponentState.cons)
        return PluginState(dispatch_config, data, plugin, component_state, messages, message_log, trans_log,
                           log_handler)

    def __init__(
            self,
            dispatch_config: DispatchConfig,
            data: D,
            plugin: Any,
            components: List[ComponentState],
            messages: PrioQueue[Message],
            message_log: List[Message],
            trans_log: List[str],
            log_handler: Maybe[logging.Handler],
    ) -> None:
        self.dispatch_config = dispatch_config
        self.data = data
        self.plugin = plugin
        self.components = components
        self.messages = messages
        self.message_log = message_log
        self.trans_log = trans_log
        self.log_handler = log_handler

    def enqueue(self, messages: List[Sendable]) -> 'PluginState[D]':
        envelopes = messages / Envelope.from_sendable
        return self.copy(messages=envelopes.fold_left(self.messages)(lambda q, m: q.put(m, m.prio)))

    def update(self, data: D) -> 'PluginState[D]':
        return self.copy(data=data)

    def log_messages(self, msgs: List[Message]) -> 'PluginState[D]':
        return self.append.message_log(msgs)

    def log_message(self, msg: Message) -> 'PluginState[D]':
        return self.log_messages(List(msg))

    def log_trans(self, trans: str) -> 'PluginState[D]':
        return self.append1.trans_log(trans)

    @property
    def has_messages(self) -> Boolean:
        return not self.messages.empty

    @property
    def unwrapped_messages(self) -> List[Message]:
        return self.messages.items / _[1] / _.msg

    @property
    def config(self) -> 'ribosome.config.Config':
        return self.dispatch_config.config

    @property
    def name(self) -> str:
        return self.config.name

    def component(self, name: str) -> Either[str, ComponentState]:
        return self.components.by_name(name)


class PluginStateHolder(Generic[D], Dat['PluginStateHolder'], Logging):

    @staticmethod
    def concurrent(state: PluginState[D], log_handler: logging.Handler=None) -> 'PluginStateHolder[D]':
        return ConcurrentPluginStateHolder(state, Maybe.check(log_handler), Lock(), False)

    @staticmethod
    def strict(state: PluginState[D], log_handler: logging.Handler=None) -> 'PluginStateHolder[D]':
        return StrictPluginStateHolder(state, log_handler)

    def __init__(self, state: PluginState[D], log_handler: Maybe[logging.Handler]=Nothing) -> None:
        self.state = state
        self.log_handler = log_handler

    def update(self, state: PluginState[D]) -> None:
        self.state = state

    @abc.abstractmethod
    def acquire(self) -> NvimIO[None]:
        ...

    @abc.abstractmethod
    def release(self, error: Optional[Exception]=None) -> NvimIO[None]:
        ...

    @abc.abstractmethod
    def dispatch_complete(self, dispatch: Dispatch) -> IO[None]:
        ...

    @property
    def has_messages(self) -> Boolean:
        return self.state.has_messages

    @property
    def message_queue(self) -> PrioQueue[Message]:
        return self.state.messages

    @property
    def messages(self) -> List[Message]:
        return self.state.unwrapped_messages


class ConcurrentPluginStateHolder(Generic[D], PluginStateHolder[D]):

    def __init__(
            self,
            state: PluginState[D],
            log_handler: Maybe[logging.Handler],
            lock: Lock,
            running: Boolean,
    ) -> None:
        super().__init__(state, log_handler)
        self.lock = lock
        self.running = running
        self.waiting_greenlet = None

    def _set_waiting_greenlet(self) -> None:
        self.waiting_greenlet = greenlet.getcurrent()

    def _unset_waiting_greenlet(self, error: Exception=None) -> None:
        if error:
            self.log.caught_exception(f'switching to parent greenlet in `acquire`', error)
        gr = self.waiting_greenlet
        self.waiting_greenlet = None
        return gr

    # FIXME if an async request arrives while another async dispatch is executing, will it behave similarly?
    # If so, this needs multiple waiting greenlets.
    # If not, the parent switch (and overwriting of `waiting_greenlet`) may not be done.
    @do(NvimIO[None])
    def acquire(self) -> Do:
        '''acquire the state lock that prevents multiple dispatches from updating the state asynchronously.
        If the lock is already acquired, an async dispatch is currently executing while another (sync or async) has been
        requested. In order not to block on requests to vim from the running dispatch, the greenlet that was started by
        the vim session must be suspended, giving control back to the running async dispatch at the point where the vim
        request was made.
        '''
        def switch() -> None:
            self.log.debug(f'acquire: switching to running dispatch')
            self._set_waiting_greenlet()
            Try(self.waiting_greenlet.parent.switch).leffect(self._unset_waiting_greenlet)
        if self.running:
            yield NvimIO.simple(switch)
        yield NvimIO.simple(self.lock.acquire)
        yield NvimIO.simple(setattr, self, 'running', True)

    @do(NvimIO[None])
    def release(self, error: Optional[Exception]=None) -> Do:
        yield NvimIO.simple(setattr, self, 'running', False)
        yield NvimIO.simple(Try, self.lock.release)
        if error:
            self.log.debug(f'released lock due to error: {error}')
            yield NvimIO.exception(error)

    @do(IO[None])
    def dispatch_complete(self, dispatch: Dispatch) -> Do:
        '''switch back to the dispatch that was suspended in `acquire` while the dispatch executing this method was
        running.
        '''
        if self.waiting_greenlet is not None:
            self.log.debug('release: switching to waiting dispatch')
            gr = yield IO.delay(self._unset_waiting_greenlet)
            yield IO.delay(gr.switch)
        yield IO.pure(None)


class StrictPluginStateHolder(Generic[D], PluginStateHolder[D]):
    pass

    def acquire(self) -> NvimIO[None]:
        return NvimIO.pure(None)

    def release(self, error: Optional[Exception]=None) -> NvimIO[None]:
        return NvimIO.pure(None)

    def dispatch_complete(self, dispatch: Dispatch) -> IO[None]:
        return IO.pure(None)


__all__ = ('ComponentState', 'PluginState', 'PluginStateHolder', 'Components')



import abc
from typing import Callable, TypeVar, Type, Generic, Any

from amino import List, Lists, L, _
from amino.dat import Dat

from ribosome.trans.effect import TransEffect, cont, lift
from ribosome.trans.action import TransAction, TransM, TransMPure, TransFailure
from ribosome.trans.message_base import Message
from ribosome.dispatch.component import Component
from ribosome.request.args import ArgValidator, ParamsSpec

D = TypeVar('D')
M = TypeVar('M', bound=Message)
R = TypeVar('R')
O = TypeVar('O')
A = TypeVar('A')
B = TypeVar('B')


class TransComplete(Dat['TransComplete']):

    def __init__(self, name: str, action: TransAction) -> None:
        self.name = name
        self.action = action


def extract(name: str, output: O, effects: List[TransEffect]) -> TransComplete:
    trans_result = cont(effects, False, lambda f: f(output)) | output
    return TransComplete(name, lift(trans_result, False))


class TransHandler(abc.ABC):

    @abc.abstractmethod
    def __call__(self, *args: Any) -> 'TransHandler':
        ...


class MessageTransHandler(Generic[M, D], Dat['MessageTransHandler[M, D]'], TransHandler):

    @staticmethod
    def create(fun: Callable[[M], R], msg: Type[M], effects: List[TransEffect], prio: float) -> 'TransHandler':
        name = fun.__name__
        return MessageTransHandler(name, fun, msg, effects, prio)

    def __init__(self, name: str, fun: Callable[[M], R], message: Type[M], effects: List[TransEffect], prio: float
                 ) -> None:
        self.name = name
        self.message = message
        self.fun = fun
        self.prio = prio
        self.effects = effects

    def run(self, component: Component, msg: M) -> TransAction:
        return extract(self.name, self.fun(component, msg), Lists.wrap(self.effects))

    def __call__(self, *args: Any) -> 'MessageTransHandler':
        return self


class FreeTransHandler(Generic[D, R], Dat['FreeTransHandler[M, D]'], TransHandler):

    @staticmethod
    def create(fun: Callable[..., R], effects: List[TransEffect], prio: float) -> 'FreeTransHandler':
        name = fun.__name__
        return FreeTransHandler(name, fun, (), effects, prio)

    def __init__(self, name: str, fun: Callable[..., R], args: tuple, effects: List[TransEffect], prio: float) -> None:
        self.name = name
        self.fun = fun
        self.args = args
        self.effects = effects
        self.prio = prio

    def __call__(self, *args: Any) -> 'FreeTransHandler':
        return self.copy(args=args)

    def execute(self) -> TransAction:
        val = ArgValidator(self.params_spec)
        return val.either(self.args, 'trans', self.name).bimap(TransFailure, lambda a: self.fun(*self.args))

    def run(self) -> TransComplete:
        return self.execute().cata(L(TransComplete)(self.name, _), L(extract)(self.name, _, Lists.wrap(self.effects)))

    @property
    def m(self) -> TransM:
        return TransMPure(self)

    @property
    def params_spec(self) -> ParamsSpec:
        return ParamsSpec.from_function(self.fun)


__all__ = ('MessageTransHandler', 'FreeTransHandler')



import abc
from typing import Any

from amino import List, Either, Right, __, Map
from amino.json.decoder import decode_json_type

from ribosome.request.args import ParamsSpec


class ArgParser(abc.ABC):

    def __init__(self, params_spec: ParamsSpec) -> None:
        self.params_spec = params_spec

    @abc.abstractmethod
    def parse(self, args: List[Any]) -> Either[str, List[Any]]:
        ...


class TokenArgParser(ArgParser):

    def parse(self, args: List[Any]) -> Either[str, List[Any]]:
        return Right(args)


class JsonArgParser(ArgParser):

    def parse(self, args: List[Any]) -> Either[str, List[Any]]:
        def parse(start: int) -> Either[str, List[Any]]:
            strict = args[:start]
            json_args = args[start:].join_tokens
            tpe = self.params_spec.types.last | (lambda: Map)
            return decode_json_type(json_args, tpe) / strict.cat
        return args.index_where(__.startswith('{')) / parse | Right(args)


__all__ = ('ArgParser', 'TokenArgParser', 'JsonArgParser')
